/*
 * Copyright 2014 Adobe Systems Incorporated (http://www.adobe.com/).
 * All Rights Reserved.
 *
 * This software is licensed as OpenSource, under the Apache License, Version
 * 2.0.
 * This license is available at: http://opensource.org/licenses/Apache-2.0.
 */

#ifndef LIBPSAUTOHINT_SRC_OPTABLE_H_
#define LIBPSAUTOHINT_SRC_OPTABLE_H_

#include "ac.h"
#include "opcodes.h"

char* GetOperator(int16_t op);

#endif /* LIBPSAUTOHINT_SRC_OPTABLE_H_ */
